/*******************************************************************************
 * (c) Copyright 2011 Microsemi Corporation.  All rights reserved.
 *
 * Marvell's M88E1111 PHY interface driver source code.
 * Ported from: linux/drivers/net/phy/marvell.c in Linux-2.6.35.3
 *
 * SVN $Revision: 4791 $
 * SVN $Date: 2012-12-06 10:20:29 +0000 (Thu, 06 Dec 2012) $
 */
#include "phy.h"
#include "m88e1111_phy.h"
#include "mss_ethernet_mac_types.h"
#include "mss_ethernet_mac.h"

#ifdef __cplusplus
extern "C" {
#endif
/**************************************************************************/
/* Preprocessor Macros                                                    */
/**************************************************************************/

#define SF2_PHY_ADDR    0x1E

#define SF2_MSGMII_PHY_RESET_MASK  0x8000u


#define SERDES_LANE3_SOFTRESET_MASK     0x20u

/**************************************************************************/
/* Private Function definitions                                           */
/**************************************************************************/



/**************************************************************************/
/* Public Function definitions                                            */
/**************************************************************************/

/**************************************************************************//**
 * 
 */
static void config_serdes(void)
{
    uint32_t temp;
    
    SERDESIF_TypeDef * SERDES0 = (SERDESIF_TypeDef *)0x40028000;
    
    /*
     * Default SGMII configuration
     */
/*     
    SERDES0->lane[3].CR0 = 0x80;
    SERDES0->lane[3].ERRCNT_DEC = 0x20;
    SERDES0->lane[3].RXIDLE_MAX_ERRCNT_THR = 0xF8;
    SERDES0->lane[3].IMPED_RATIO = 0x80;
    SERDES0->lane[3].PLL_F_PCLK_RATIO = 0x00;
    SERDES0->lane[3].PLL_M_N = 0x29;
    SERDES0->lane[3].CNT250NS_MAX = 0x20;
    SERDES0->lane[3].RE_AMP_RATIO = 0x00;
    SERDES0->lane[3].RE_CUT_RATIO = 0x00;
    SERDES0->lane[3].TX_AMP_RATIO = 0x80;
    SERDES0->lane[3].TX_PST_RATIO = 0x00;
    SERDES0->lane[3].TX_PRE_RATIO = 0x00;
    SERDES0->lane[3].ENDCALIB_MAX = 0x10;
    SERDES0->lane[3].CALIB_STABILITY_COUNT = 0x38;
    SERDES0->lane[3].POWER_DOWN = 0x00;
    SERDES0->lane[3].RX_OFFSET_COUNT = 0x70;
    SERDES0->lane[3].PLL_F_PCLK_RATIO_5GBPS = 0x02;
    SERDES0->lane[3].PLL_M_N_5GBPS = 0x22;
*/    
    /*
     * Modified configuration for SGMII
     */
#if 0     
    SERDES0->lane[3].GEN1_RX_PLL_CCP = 0x00000022;
    SERDES0->lane[3].GEN1_TX_PLL_CCP = 0x00000002;
    SERDES0->lane[3].TX_PST_RATIO = 0x00000015;
    SERDES0->lane[3].CNT250NS_MAX = 0x00000020;
    SERDES0->lane[3].PLL_M_N = 0x00000029;
    SERDES0->lane[3].ENDCALIB_MAX  = 0x00000000;
    SERDES0->sys_regs.CONFIG_PHY_MODE_0 = 0x00004444;
//    *(PCIE_CORECLK_SEL) = 0x00000030;
    SERDES0->sys_regs.PCLK_SEL = 0x00000030;
    SERDES0->lane[3].TX_PRE_RATIO = 0x00000080;
    SERDES0->lane[3].UPDATE_SETTINGS = 0x00000001;
    
    /* Assert then de-assert SERDES lane3 reset. */
    SERDES0->sys_regs.SERDESIF_SOFT_RESET |= SERDES_LANE3_SOFTRESET_MASK;
    SERDES0->sys_regs.SERDESIF_SOFT_RESET &= ~SERDES_LANE3_SOFTRESET_MASK;
#endif    
    // From Jagannath's validation design:
    SERDES0->lane[3].GEN1_RX_PLL_CCP = 0x00000022;
    SERDES0->lane[3].GEN1_TX_PLL_CCP = 0x00000002;

    SERDES0->lane[3].CNT250NS_MAX = 0x00000020;
    SERDES0->lane[3].PLL_M_N =  0x00000029;
    SERDES0->lane[3].ENDCALIB_MAX = 0x00000010;
    SERDES0->sys_regs.CONFIG_PHY_MODE_0 = 0x00004444;
    SERDES0->sys_regs.PCLK_SEL = 0x00000030;
    SERDES0->lane[3].TX_PST_RATIO = 0x00000000;
    SERDES0->lane[3].CNT250NS_MAX = 0x00000020;
    SERDES0->lane[3].PLL_M_N = 0x00000029;
    SERDES0->lane[3].ENDCALIB_MAX = 0x00000000;
    SERDES0->sys_regs.CONFIG_PHY_MODE_0 = 0x00004444;
    SERDES0->lane[3].UPDATE_SETTINGS = 0x00000001;
    SERDES0->sys_regs.EPCS_RSTN_SEL = 0x00000001;
    SERDES0->lane[3].TX_PRE_RATIO = 0x00000000;
    SERDES0->lane[3].UPDATE_SETTINGS = 0x00000001;
    
    
    
    /* Assert then de-assert PHY reset. */
    temp = MSS_MAC_read_phy_reg(SF2_PHY_ADDR, 0x00);
    temp |= SF2_MSGMII_PHY_RESET_MASK;
    MSS_MAC_write_phy_reg(SF2_PHY_ADDR, 0x00, temp);
    temp &= ~SF2_MSGMII_PHY_RESET_MASK;
    MSS_MAC_write_phy_reg(SF2_PHY_ADDR, 0x00, temp);
}

/**************************************************************************//**
 * See m88e1111_phy.h for details of how to use this function.
 */
return_t m88e1111_init(MAC_cfg_t * cfg)
{
    volatile uint16_t msgmii_control;
    volatile uint16_t msgmii_status;
    volatile uint16_t msgmii_an_adv;
    volatile uint16_t msgmii_partner_ability;
    volatile uint16_t msgmii_an_expansion;
    volatile uint16_t msgmii_an_next_page_tx;
    volatile uint16_t msgmii_partner_next;
    volatile uint16_t msgmii_extended_sr;
    volatile uint16_t msgmii_tbi_cr;

    uint16_t phy_reset_in_progress;
    
	uint32_t err;
	uint32_t temp;

	/* Enable Fiber/Copper auto selection */
/*    
	temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, M88E1111_PHY_EXT_SR);
	temp &= ~M88E1111_HWCFG_FIBER_COPPER_AUTO;
	MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, M88E1111_PHY_EXT_SR, temp);

	temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, MII_BMCR);
	temp |= BMCR_RESET;
	MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, MII_BMCR, temp);
*/
	if ((cfg->interface == RGMII))
    {
    	temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_CR);
		if (temp == 0)
			return temp;

		err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_CR, temp);

		if (err == 0)
			return err;

		temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_SR);
		if (temp == 0)
			return temp;

		temp &= ~(M88E1111_HWCFG_MODE_MASK);

		if (temp & M88E1111_HWCFG_FIBER_COPPER_RES)
			temp |= M88E1111_HWCFG_MODE_FIBER_RGMII;
		else
			temp |= M88E1111_HWCFG_MODE_COPPER_RGMII;

		err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_SR, temp);
		if (err == 0)
			return err;
	}
    else if(TBI == cfg->interface)
    {
        config_serdes();
    
        {
            /*
             * Enable auto-negotiation inside SmartFusion2 SGMII block.
             */
            temp = MSS_MAC_read_phy_reg(0x1E, 0x00);
                                        
            temp |= 0x1000;
                                        
            err = MSS_MAC_write_phy_reg(0x1E, 0x00, temp);
            
            MSS_MAC_write_phy_reg(0x1E, 0x04, 0x0001);
                                        
            msgmii_control = MSS_MAC_read_phy_reg(0x1E, 0x00);
            msgmii_status = MSS_MAC_read_phy_reg(0x1E, 0x01);
            msgmii_an_adv = MSS_MAC_read_phy_reg(0x1E, 0x04);
            msgmii_partner_ability = MSS_MAC_read_phy_reg(0x1E, 0x05);
            msgmii_an_expansion = MSS_MAC_read_phy_reg(0x1E, 0x06);
            msgmii_an_next_page_tx = MSS_MAC_read_phy_reg(0x1E, 0x07);
            msgmii_partner_next = MSS_MAC_read_phy_reg(0x1E, 0x08);
            msgmii_extended_sr = MSS_MAC_read_phy_reg(0x1E, 0x0F);
            msgmii_tbi_cr = MSS_MAC_read_phy_reg(0x1E, 0x11);

            ++msgmii_control;
            ++msgmii_status;
            ++msgmii_an_adv;
            ++msgmii_partner_ability;
            ++msgmii_an_expansion;
            ++msgmii_an_next_page_tx;
            ++msgmii_partner_next;
            ++msgmii_extended_sr;
            ++msgmii_tbi_cr;
        }
        /**/
        err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, M88E1111_EXT_ADDR_PAGE_CR, PAGE_0);

                                    

        temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, M88E1111_CR);
        temp = 0x9340;
        err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, M88E1111_CR, temp);
                                    

        temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, M88E1111_EXT_ADDR_PAGE_CR);
        
        err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, M88E1111_EXT_ADDR_PAGE_CR, PAGE_0);
                                    
        
        temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, MII_BMCR);
        temp |= BMCR_RESET;
        MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, MII_BMCR, temp);
        do {
            temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, MII_BMCR);
            phy_reset_in_progress = temp & BMCR_RESET;
        } while(phy_reset_in_progress);

        
        
        {
            temp = MSS_MAC_read_phy_reg(0x1E, 0x00);
            
            if (temp == 0) { return temp; }
                                        
            temp |= 0x0200;
                                        
            err = MSS_MAC_write_phy_reg(0x1E, 
                                        0x00, temp);

            msgmii_control = MSS_MAC_read_phy_reg(0x1E, 0x00);
            msgmii_status = MSS_MAC_read_phy_reg(0x1E, 0x01);
            msgmii_an_adv = MSS_MAC_read_phy_reg(0x1E, 0x04);
            msgmii_partner_ability = MSS_MAC_read_phy_reg(0x1E, 0x05);
            msgmii_an_expansion = MSS_MAC_read_phy_reg(0x1E, 0x06);
            msgmii_an_next_page_tx = MSS_MAC_read_phy_reg(0x1E, 0x07);
            msgmii_partner_next = MSS_MAC_read_phy_reg(0x1E, 0x08);
            msgmii_extended_sr = MSS_MAC_read_phy_reg(0x1E, 0x0F);
            msgmii_tbi_cr = MSS_MAC_read_phy_reg(0x1E, 0x11);

            msgmii_status = MSS_MAC_read_phy_reg(0x1E, 0x01);
            
            ++msgmii_control;
            ++msgmii_status;
            ++msgmii_an_adv;
            ++msgmii_partner_ability;
            ++msgmii_an_expansion;
            ++msgmii_an_next_page_tx;
            ++msgmii_partner_next;
            ++msgmii_extended_sr;
            ++msgmii_tbi_cr;
        }
        
        
        
        /* Read status*/
/*
        temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, 0x11);
                                    
        if (temp == 0) { return temp; }

        err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 0x11, temp);
*/        
        
        
        
        /**/
/*
        temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_CR);
        
        if (temp == 0) { return temp; }

        err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_CR, temp);
*/
/*                                    
        temp = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_SR);
        if (temp == 0) { return temp; }

        temp &= ~(M88E1111_HWCFG_MODE_MASK);

        if (temp & M88E1111_HWCFG_FIBER_COPPER_RES)
            temp |= M88E1111_HWCFG_MODE_FIBER_RGMII;
        else
            temp |= M88E1111_HWCFG_MODE_COPPER_RGMII;

        err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                    M88E1111_PHY_EXT_SR, temp);
        if (err == 0)
            return err;
*/
    }
	/* SGMII support code is removed. Because G4M MAC 
       internally redirects to inbuilt SGMII PHY when 
       SGMII interface is chosen */
    else
        return 0; /* Unsupported interface */
/*    
    err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                MII_BMCR, BMCR_RESET);
	if (err == 0)
		return err;
*/
    {
        temp = MSS_MAC_read_phy_reg(0x1E, 0x01);
        
        if (temp == 0) { return temp; }
    }
	return 1;
}

/**************************************************************************//**
 * See m88e1111_phy.h for details of how to use this function.
 */
return_t m88e1111_config_aneg(MAC_cfg_t * cfg)
{
	uint32_t err;
#if 0
	/* The Marvell PHY has an errata which requires
	 * that certain registers get written in order
	 * to restart autonegotiation */
	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                MII_BMCR, BMCR_RESET);

	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                0x1d, 0x1f);
	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                0x1e, 0x200c);
	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                0x1d, 0x5);
	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                0x1e, 0);
	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                0x1e, 0x100);
	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                M88E1111_PHY_SCR,
			                    M88E1111_PHY_SCR_AUTO_CROSS);
	if (err == 0)
		return err;

	err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR,
                                M88E1111_PHY_LED_CONTROL,
			                    M88E1111_PHY_LED_DIRECT);
	if (err == 0)
		return err;

	err = phy_config_aneg(M88E1111, cfg);
	if (err == 0)
		return err;

	if (cfg->auto_negotiation != MSS_MAC_AUTO_NEGOTIATION_ENABLE) 
    {
		int bmcr;

		/*
		 * A write to speed/duplex bits (that is performed by
		 * genphy_config_aneg() call above) must be followed by
		 * a software reset. Otherwise, the write has no effect.
		 */
		bmcr = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, MII_BMCR);
		if (bmcr == 0)
			return bmcr;

		err = MSS_MAC_write_phy_reg(M88E1111_PHY_ADDR, 
                                    MII_BMCR, bmcr | BMCR_RESET);
		if (err == 0)
			return err;
	}
#endif
    /* Auto-negotiation successful */
    return 1;
}

/**************************************************************************//**
 * See m88e1111_phy.h for details of how to use this function.
 */
return_t m88e1111_read_status(MAC_cfg_t * cfg)
{
	uint32_t adv;
	uint32_t err;
	uint32_t lpa;
	uint32_t status = 0;

	/* Update the link, but return if there
	 * was an error */
	err = phy_update_link( M88E1111 );
	if (!err)
		return err;
    /* Proceed further only if the link is UP */
	if (MSS_MAC_AUTO_NEGOTIATION_ENABLE == cfg->auto_negotiation) 
    {
		status = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR,
                                      M88E1111_PHY_STATUS);
		if (status == 0)
			return status;

		lpa = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR,
                                   MII_LPA);
		if (lpa == 0)
			return lpa;

		adv = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR, 
                                   MII_ADVERTISE);
		if (adv == 0)
			return adv;

		lpa &= adv;

		if (status & M88E1111_PHY_STATUS_FULLDUPLEX)
			cfg->fullduplex = MSS_MAC_FULLDUPLEX_ENABLE;
		else
			cfg->fullduplex = MSS_MAC_FULLDUPLEX_DISABLE;

		status = status & M88E1111_PHY_STATUS_SPD_MASK;

		switch (status) 
        {
    		case M88E1111_PHY_STATUS_1000:
    			cfg->speed = MAC1000MBPS;
    			break;
    
    		case M88E1111_PHY_STATUS_100:
    			cfg->speed = MAC100MBPS;
    			break;
    
    		default:
    			cfg->speed = MAC10MBPS;
    			break;
		}
        if( cfg->fullduplex == MSS_MAC_FULLDUPLEX_ENABLE)
        {
            //Configure pause and asymmetric pause settings for PHY.            
        }
	} 
    else 
    {
		uint32_t bmcr = MSS_MAC_read_phy_reg(M88E1111_PHY_ADDR,
                                             MII_BMCR);
		if (bmcr == 0)
			return bmcr;

		if (bmcr & BMCR_FULLDPLX)
			cfg->fullduplex = MSS_MAC_FULLDUPLEX_ENABLE;
		else
			cfg->fullduplex = MSS_MAC_FULLDUPLEX_DISABLE;

		if (bmcr & BMCR_SPEED1000)
			cfg->speed = MAC1000MBPS;
		else if (bmcr & BMCR_SPEED100)
			cfg->speed = MAC100MBPS;
		else
			cfg->speed = MAC10MBPS;
	}
	return 1;
}

#ifdef __cplusplus
}
#endif

/******************************** END OF FILE ******************************/






